// This file is part of nullboot
// Copyright 2021 Canonical Ltd.
// SPDX-License-Identifier: GPL-3.0-only

// Package nullboot provides the nullboot experience
// Most of it is in subpackages.
package nullboot
